DP_CHECKPOINT_FILE = 'model.ckpt'
DP_DISP_FILE = 'lcurve.out'
DP_PROFILING_FILE = 'timeline.json'
DP_INPUT_FILE = 'input.json'
DP_FROZEN_MODEL = 'frozen_model.pb'
DP_ORIGINAL_MODEL = 'original_model.pb'

MODEL_DEVI_OUT = 'model_devi.out'
MODEL_DEVI_NEU_OUT = 'model_devi_neu.out'
MODEL_DEVI_RED_OUT = 'model_devi_red.out'

LAMMPS_TRAJ_DIR = 'traj'
LAMMPS_TRAJ_SUFFIX = '.lammpstrj'


from ai2_kit.core.artifact import Artifact
from ai2_kit.core.script import BashTemplate
from ai2_kit.core.script import BashScript, BashStep
from ai2_kit.core.job import JobFuture, gather_jobs
from ai2_kit.core.log import get_logger

from pydantic import BaseModel
from typing import List
from dataclasses import dataclass
import os
import copy
import random
import sys
import json

from .cll import ICllTrainOutput, BaseCllContext
from .data_helper import Cp2kOutputHelper, DeepmdNpyHelper, convert_to_deepmd_npy
from .constant import (
    DP_CHECKPOINT_FILE,
    DP_DISP_FILE,
    DP_PROFILING_FILE,
    DP_INPUT_FILE,
    DP_FROZEN_MODEL,
    DP_ORIGINAL_MODEL
)

logger = get_logger(__name__)


class GenericDeepmdInputConfig(BaseModel):
    model_num: int = 4
    init_dataset: List[str]
    input_template: dict
    compress_model: bool = False

class GenericDeepmdContextConfig(BaseModel):
    script_template: BashTemplate
    dp_cmd: str = 'dp'


@dataclass
class GenericDeepmdInput:
    config: GenericDeepmdInputConfig
    type_map: List[str]
    old_dataset: List[Artifact]  # training data used by previous iteration
    new_dataset: List[Artifact]  # training data used by current iteration
    initiated: bool = False

    def update_training_dataset(self, dataset: List[Artifact]):
        self.old_dataset += self.new_dataset
        self.new_dataset = dataset


@dataclass
class GenericDeepmdContext(BaseCllContext):
    config: GenericDeepmdContextConfig


@dataclass
class GenericDeepmdOutput(ICllTrainOutput):
    outputs: List[Artifact]
    input: GenericDeepmdInput

    def get_mlp_models(self) -> List[Artifact]:
        return [a.join(DP_FROZEN_MODEL) for a in self.outputs]

    def get_training_dataset(self) -> List[Artifact]:
        return self.input.new_dataset + self.input.old_dataset


async def generic_deepmd(input: GenericDeepmdInput, ctx: GenericDeepmdContext):
    executor = ctx.resource_manager.default_executor

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [converted_data_dir, tasks_dir] = executor.setup_workspace(
        work_dir, ['converted_input_data', 'tasks'])

    # initialization
    if not input.initiated:
        input.new_dataset += ctx.resource_manager.resolve_artifacts(
            input.config.init_dataset)

    # convert data type if necessary, only needed for new data as old data is already converted
    new_deepmd_npy_data: List[Artifact] = []
    cp2k_output_data: List[Artifact] = []

    # TODO: refactor data type conversion
    for artifact in input.new_dataset:
        if not artifact.format or DeepmdNpyHelper.is_match(artifact):
            # treated as deepmd npy data if format is not specified
            new_deepmd_npy_data.append(artifact)
        elif Cp2kOutputHelper.is_match(artifact):
            cp2k_output_data.append(artifact)
        else:
            raise ValueError(f'unsupported data type: {artifact.format}')

    # convert data to deepmd/npy format
    # TODO: support more data type
    converted_data_dirs = executor.run_python_fn(convert_to_deepmd_npy)(
        cp2k_outputs=[a.to_dict() for a in cp2k_output_data],
        base_dir=converted_data_dir,
        type_map=input.type_map,
    )
    new_deepmd_npy_data += [Artifact.of(
        url=url, format=DeepmdNpyHelper.format,
    ) for url in converted_data_dirs]

    input.new_dataset = new_deepmd_npy_data

    jobs: List[JobFuture] = []
    output_dirs = []

    # train multiple models at the same time
    for i in range(input.config.model_num):
        # each model should be trained in its own task_dir
        task_dir = os.path.join(tasks_dir, str(i).zfill(3))
        executor.mkdir(task_dir)

        # create dp train input file
        # NOTE: dp v1 format is supported currently
        # TODO: support more params if it is necessary

        # ref: https://github.com/deepmodeling/dpgen2/blob/master/examples/ch4/param_CH4_deepmd-kit-2.1.1.json
        # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/prep_dp_train.py
        # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/run_dp_train.py

        dp_input = copy.deepcopy(input.config.input_template)
        training: dict = dp_input['training']

        # set output files
        training['disp_file'] = DP_DISP_FILE
        training['save_ckpt'] = DP_CHECKPOINT_FILE
        training['profiling_file'] = DP_PROFILING_FILE

        # set random seed
        discriptor = dp_input['model']['descriptor']
        if discriptor['type'] == 'hybrid':
            for d in discriptor['list']:
                d['seed'] = _random_seed()
        else:
            discriptor['seed'] = _random_seed()
        dp_input['model']['fitting_net']['seed'] = _random_seed()
        dp_input['training']['seed'] = _random_seed()

        # set training data
        systems = [a.url for a in input.old_dataset + input.new_dataset]
        training['systems'] = systems
        set_prefix: str = training.setdefault(
            'set_prefix', 'set')  # respect user input
        auto_prob_str = "prob_sys_size"
        training.setdefault('batch_size', 'auto')
        training['auto_prob_style'] = auto_prob_str

        # v2 training data
        training_data = {
            'systems': training['systems'],
            'set_prefix': training['set_prefix'],
            'auto_prob_style': training['auto_prob_style'],
            'batch_size': training['batch_size'],
        }
        training['training_data'] = training_data

        # other params
        dp_input['model']['type_map'] = input.type_map

        # write config to executor
        dp_input_text = json.dumps(dp_input, indent=2)
        dp_input_path = os.path.join(task_dir, DP_INPUT_FILE)
        executor.dump_text(dp_input_text, dp_input_path)

        # build script
        dp_cmd = ctx.config.dp_cmd
        dp_train_cmd = [dp_cmd, 'train', DP_INPUT_FILE]

        steps = [
            BashStep(cmd=dp_train_cmd, checkpoint='dp-train') # type: ignore
        ]

        if input.config.compress_model:
            dp_freeze_cmd = [dp_cmd, 'freeze', '-o', DP_ORIGINAL_MODEL]
            dp_compress_cmd = \
                [dp_cmd, 'compress', '-i', DP_ORIGINAL_MODEL, '-o', DP_FROZEN_MODEL]
            steps.append(BashStep(cmd=dp_freeze_cmd)) # type: ignore
            steps.append(BashStep(cmd=dp_compress_cmd)) # type: ignore
        else:
            dp_freeze_cmd = [dp_cmd, 'freeze', '-o', DP_FROZEN_MODEL]
            steps.append(BashStep(cmd=dp_freeze_cmd)) # type: ignore

        dp_train_script = BashScript(
            template=ctx.config.script_template,
            steps=steps # type: ignore
        )
        output_dirs.append(task_dir)

        # submit job
        job = executor.submit(dp_train_script.render(), cwd=task_dir, checkpoint_key=f'submit-job/dp-train/{i}:{task_dir}')
        jobs.append(job)

    await gather_jobs(jobs, max_tries=2)

    return GenericDeepmdOutput(
        input=input,
        outputs=[Artifact.of(
            url=url,
        ) for url in output_dirs]
    )


def _random_seed():
    return random.randrange(sys.maxsize) % (1 << 32)



from ai2_kit.core.script import BashTemplate, BashStep, BashScript
from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.log import get_logger
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.util import split_list, dict_nested_get

from typing import List, Literal, Optional, Mapping, Sequence, Any
from pydantic import BaseModel
from dataclasses import dataclass
from string import Template
from allpairspy import AllPairs
import os
import itertools
import random


from .cll import BaseCllContext, ICllExploreOutput
from .constant import (
    MODEL_DEVI_OUT,
    MODEL_DEVI_NEU_OUT,
    MODEL_DEVI_RED_OUT,
    LAMMPS_TRAJ_DIR,
    LAMMPS_TRAJ_SUFFIX,
)
from .data_helper import LammpsOutputHelper, PoscarHelper, convert_to_lammps_input_data

logger = get_logger(__name__)


class ExploreVariants(BaseModel):
    temp: List[float]
    """Temperatures variants."""
    pres: List[float]
    """Pressures variants."""
    others: Mapping[str, Sequence[Any]] = dict()
    """
    Other variants to be combined with.
    The key is the name of the variant, and the value is the list of values.
    For example, if you want to combine the variant 'LAMBDA' with values [0.0, 0.5, 1.0],
    you can set the others field to {'LAMBDA': [0.0, 0.5, 1.0]}.
    And in LAMMPS input template, you can use the variable ${LAMBDA} and v_LAMBDA to access the value.
    """


class GenericLammpsInputConfig(BaseModel):
    explore_vars: ExploreVariants
    """Variants to be explored."""

    n_wise: int = 0
    """The way of combining variants.
    0 means cartesian product, 2 means 2-wise, etc.
    If n_wise is less than 2 or greater than total fields,
    the full combination will be used.
    It is strongly recommended to use n_wise when the full combination is too large.
    """
    system_files: List[str]
    """Artifacts of initial system data."""
    plumed_config: Optional[str]
    """Plumed config file content."""
    plumed_config_file: Optional[str]
    """Plumed config file path."""

    # ensemble specific params
    tau_t: float = 0.1
    tau_p: float = 0.5
    time_const: float = 0.1
    ensemble: Literal['nvt', 'nvt-i', 'nvt-a', 'nvt-iso', 'nvt-aniso', 'npt', 'npt-t', 'npt-tri', 'nve', 'csvr']

    no_pbc: bool = False
    timestep: float = 0.0005
    sample_freq: int
    nsteps: int

    """Ensemble to be used.
    nvt means constant volume and temperature.
    nvt-i means constant volume and temperature, with isotropic scaling.
    nvt-a means constant volume and temperature, with anisotropic scaling.
    nvt-iso means constant volume and temperature, with isotropic scaling.
    npt means constant pressure and temperature.
    npt-t means constant pressure and temperature, with isotropic scaling.
    npt-tri means constant pressure and temperature, with anisotropic scaling.
    nve means constant energy.
    """

    input_template: Optional[str]
    """Lammps input template file content."""

    post_variables_section: str = ''
    post_init_section: str = ''
    post_read_data_section: str = ''
    post_force_field_section: str = ''
    post_md_section: str = ''
    post_run_section: str = ''


class GenericLammpsContextConfig(BaseModel):
    script_template: BashTemplate
    lammps_cmd: str = 'lmp'
    concurrency: int = 5


@dataclass
class GenericLammpsInput:

    @dataclass
    class MdOptions:
        models: List[Artifact]

    @dataclass
    class FepOptions:
        red_models: List[Artifact]
        neu_models: List[Artifact]

    config: GenericLammpsInputConfig
    type_map: List[str]
    mass_map: List[float]

    # The following options are mutex
    md_options: Optional[MdOptions] = None
    fep_options: Optional[FepOptions] = None


@dataclass
class GenericLammpsContext(BaseCllContext):
    config: GenericLammpsContextConfig


@dataclass
class GenericLammpsOutput(ICllExploreOutput):
    model_devi_outputs: List[Artifact]

    def get_model_devi_dataset(self) -> List[Artifact]:
        return self.model_devi_outputs


async def generic_lammps(input: GenericLammpsInput, ctx: GenericLammpsContext):
    executor = ctx.resource_manager.default_executor

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    input_data_dir, tasks_dir = executor.setup_workspace(work_dir, ['input_data', 'tasks'])

    systems = ctx.resource_manager.resolve_artifacts(input.config.system_files)

    # prepare lammps input data
    # TODO: refactor the way of handling different types of input
    # TODO: handle more data format, for example, cp2k output
    poscar_files: List[Artifact] = []
    for system_file in systems:
        if PoscarHelper.is_match(system_file):
            poscar_files.append(system_file)
        else:
            raise ValueError(f'unsupported system file type: {system_file}')

    input_data_files: List[ArtifactDict] = executor.run_python_fn(convert_to_lammps_input_data)(
        poscar_files=[a.to_dict() for a in poscar_files],
        base_dir=input_data_dir,
        type_map=input.type_map,
    )

    combination_fields: List[str] = [
        'data_file',
        'temp',
        'pres'
    ]
    combination_values: Sequence[Sequence[Any]] = [
        input_data_files,
        input.config.explore_vars.temp,
        input.config.explore_vars.pres,
    ]

    for k, v in input.config.explore_vars.others.items():
        combination_fields.append(k)
        combination_values.append(v)  # type: ignore

    if 1 < input.config.n_wise <= len(combination_fields):
        logger.info('using %d-wise combination', input.config.n_wise)
        combinations = AllPairs(combination_values, n=input.config.n_wise)
    else:
        logger.info('using full combination')
        combinations = itertools.product(*combination_values)

    lammps_task_dirs: List[ArtifactDict] = []
    lammps_input_file_name = 'lammps.input'

    for i, combination in enumerate(combinations):
        data_file: ArtifactDict = combination[0]
        temp, pres = combination[1:3]
        others_dict = dict(zip(combination_fields[3:], combination[3:]))
        lammps_task_dir = os.path.join(tasks_dir, f'{i:06d}')
        executor.mkdir(os.path.join(lammps_task_dir, LAMMPS_TRAJ_DIR))  # create dump directory for lammps or else will get error

        if input.md_options:
            force_field_section = make_md_force_field_section(
                models=[a.url for a in input.md_options.models],
            )
        elif input.fep_options:
            if 'LAMBDA_f' not in others_dict:
                raise ValueError('LAMBDA_f must be set when using FEP mode!')
            # inject the following variables for FEP mode
            others_dict['LAMBDA_i'] = '1-v_LAMBDA_f' # should not have space, or you must quote
            others_dict['plus'] = 1
            others_dict['minus'] = -1
            force_field_section = make_fep_force_field_section(
                neu_models=[a.url for a in input.fep_options.neu_models],
                red_models=[a.url for a in input.fep_options.red_models],
            )
        else:
            raise ValueError('one and only one of md_options or fep_options must be set')

        # Config plumed if either plumed_config_file or plumed_config is set.
        # The plumed config can be set in both workflow config and input data file.
        # The config in input data file has higher priority.
        plumed_config_file = dict_nested_get(data_file, ['attrs', 'lammps', 'plumed_config_file'],
                                             input.config.plumed_config_file)
        plumed_config = dict_nested_get(data_file, ['attrs', 'lammps', 'plumed_config'],
                                        input.config.plumed_config)
        if plumed_config_file is None and isinstance(plumed_config, str):
            plumed_config_file = 'plumed.input'
            plumed_file_path = os.path.join(lammps_task_dir, plumed_config_file)
            logger.info(f'found plumed config, generate {plumed_file_path}')
            executor.dump_text(plumed_config, plumed_file_path)

        template = input.config.input_template or  DEFAULT_LAMMPS_INPUT_TEMPLATE
        input_text = make_lammps_input(data_file=data_file['url'],
                                       nsteps=input.config.nsteps,
                                       timestep=input.config.timestep,
                                       trj_freq=input.config.sample_freq,
                                       temp=temp,
                                       pres=pres,
                                       tau_t=input.config.tau_t,
                                       tau_p=input.config.tau_p,
                                       time_const=input.config.time_const,
                                       ensemble=input.config.ensemble,
                                       mass_map=input.mass_map,
                                       others_dict=others_dict,
                                       force_field_section=force_field_section,
                                       template=template,
                                       post_variables_section=input.config.post_variables_section,
                                       post_init_section=input.config.post_init_section,
                                       post_read_data_section=input.config.post_read_data_section,
                                       post_force_field_section=input.config.post_force_field_section,
                                       post_md_section=input.config.post_md_section,
                                       post_run_section=input.config.post_run_section,
                                       plumed_config_file=plumed_config_file,  # type: ignore
                                       no_pbc=False,
                                       rand_start=1_000_000)
        input_file_path = os.path.join(lammps_task_dir, lammps_input_file_name)
        logger.info(f'generate lammps config {input_file_path}')

        executor.dump_text(input_text, input_file_path)
        lammps_task_dirs.append({'url': lammps_task_dir, 'attrs': data_file['attrs']})  # type: ignore

    # build scripts and submit
    lammps_cmd = ctx.config.lammps_cmd
    base_cmd = f'{lammps_cmd} -i {lammps_input_file_name}'
    cmd = f'''if [ -f md.restart.* ]; then {base_cmd} -v restart 1; else {base_cmd} -v restart 0; fi'''

    # generate steps
    steps = []
    for lammps_task_dir in lammps_task_dirs:
        steps.append(BashStep(cwd=lammps_task_dir['url'], cmd=cmd, checkpoint='lammps'))

    # submit jobs by the number of concurrency
    jobs = []
    for i, steps_group in enumerate(split_list(steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps_group,
        )
        job = executor.submit(script.render(), cwd=tasks_dir,
                              checkpoint_key=f'submit-job/lammps/{i}:{tasks_dir}')
        jobs.append(job)

    await gather_jobs(jobs, max_tries=2)

    outputs = [
        Artifact.of(
            url=task_dir['url'],
            executor=executor.name,
            format=LammpsOutputHelper.format,
            attrs=task_dir['attrs'],
        ) for task_dir in lammps_task_dirs]  # type: ignore

    return GenericLammpsOutput(model_devi_outputs=outputs)


def make_md_force_field_section(models: List[str]):
    deepmd_args = ""
    settings = [
        'pair_style deepmd %s out_freq ${THERMO_FREQ} out_file %s %s' % (' '.join(models), MODEL_DEVI_OUT, deepmd_args),
        'pair_coeff * *',
    ]
    return settings


def make_fep_force_field_section(neu_models: List[str], red_models: List[str]):
    deepmd_args = ""
    settings = [
        'pair_style hybrid/overlay &',
        '           deepmd %s out_freq ${THERMO_FREQ} out_file %s %s &' %(' '.join(neu_models), MODEL_DEVI_NEU_OUT, deepmd_args),
        '           deepmd %s out_freq ${THERMO_FREQ} out_file %s %s'   %(' '.join(red_models), MODEL_DEVI_RED_OUT, deepmd_args),
        'pair_coeff  * * deepmd 1 *',
        'pair_coeff  * * deepmd 2 *',
        '',
        'fix sampling_PES all adapt 0 &',
        '    pair deepmd:1 scale * * v_LAMBDA_f &',
        '    pair deepmd:2 scale * * v_LAMBDA_i &',
        '    scale yes',
    ]
    return settings


def make_lammps_input(data_file: str,
                      nsteps: int,
                      timestep: float,
                      trj_freq: int,
                      temp: float,
                      pres: float,
                      tau_t: float,
                      tau_p: float,
                      ensemble: str,
                      mass_map: List[float],
                      others_dict: Mapping[str, Any],

                      template: str,
                      post_variables_section: str,
                      post_init_section: str,
                      post_read_data_section: str,
                      post_force_field_section: str,
                      post_md_section: str,
                      post_run_section: str,

                      force_field_section: List[str],
                      plumed_config_file: Optional[str] = None,
                      time_const = 0.1,
                      no_pbc=False,
                      rand_start=1_000_000,
                      ):

    # FIXME: I am not sure if it is a good idea to fix it automatically
    # maybe we should consider raise an error here
    if not ensemble.startswith('npt'):
        pres = -1

    variables = [
        '# required variables',
        'variable NSTEPS      equal %d' % nsteps,
        'variable THERMO_FREQ equal %d' % trj_freq,
        'variable DUMP_FREQ   equal %d' % trj_freq,
        'variable TEMP        equal %f' % temp,
        'variable PRES        equal %f' % pres,
        'variable TAU_T       equal %f' % tau_t,
        'variable TAU_P       equal %f' % tau_p,
        'variable TIME_CONST  equal %f' % time_const,
        '',
        '# custom variables (if any)',
    ]

    for k, v in others_dict.items():
        variables.append(f'variable {k} equal {v}')

    init_section = [
        'boundary ' + ('f f f' if no_pbc else 'p p p'),
    ]

    read_data_section = [
        '''if "${restart} > 0" then "read_restart md.restart.*" else "read_data %s"''' % data_file,
        *("mass {id} {mass}".format(id=i+1, mass=m) for i, m in enumerate(mass_map))
    ]

    md_section = [
        '''if "${restart} == 0" then "velocity all create ${TEMP} %d"''' % (random.randrange(rand_start - 1) + 1)
    ]

    if ensemble.startswith('npt') and no_pbc:
        raise ValueError('ensemble npt conflict with no_pcb')
    if ensemble in ('npt', 'npt-i', 'npt-iso',):
        md_section.append('fix 1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}')
    elif ensemble in ('npt-a', 'npt-aniso',):
        md_section.append('fix 1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P}')
    elif ensemble in ('npt-t', 'npt-tri',):
        md_section.append('fix 1 all npt temp ${TEMP} ${TEMP} ${TAU_T} tri ${PRES} ${PRES} ${TAU_P}')
    elif ensemble in ('nvt',):
        md_section.append('fix 1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}')
    elif ensemble in ('nve',):
        md_section.append('fix 1 all nve')
    elif ensemble in ('csvr',):
        md_section.append('fix 1 all nve')
        md_section.append('fix 2 all temp/csvr ${TEMP} ${TEMP} ${TIME_CONST} %d' % (random.randrange(rand_start - 1) + 1))
    else:
        raise ValueError('unknown ensemble: ' + ensemble)

    if plumed_config_file:
        md_section.append(f'fix dpgen_plm all plumed plumedfile {plumed_config_file} outfile plumed.out')

    if no_pbc:
        md_section.extend([
            'velocity all zero linear',
            'fix      fm all momentum 1 linear 1 1 1',
        ])

    md_section.extend([
        'thermo_style custom step temp pe ke etotal press vol lx ly lz xy xz yz',
        'thermo       ${THERMO_FREQ}',
        'dump         1 all custom ${DUMP_FREQ} %s/*%s id type x y z fx fy fz' % (LAMMPS_TRAJ_DIR, LAMMPS_TRAJ_SUFFIX),
        'restart      10000 md.restart',
    ])

    run_section = [
        'timestep %f' % timestep,
        'run      ${NSTEPS} upto',
    ]

    return LammpsInputTemplate(template).substitute(dict(
        variables_section='\n'.join(variables),
        post_variables_section=post_variables_section,

        init_section='\n'.join(init_section),
        post_init_section=post_init_section,

        read_data_section='\n'.join(read_data_section),
        post_read_data_section=post_read_data_section,

        md_section='\n'.join(md_section),
        post_md_section=post_md_section,

        force_field_section='\n'.join(force_field_section),
        post_force_field_section=post_force_field_section,

        run_section='\n'.join(run_section),
        post_run_section=post_run_section,
    ))


class LammpsInputTemplate(Template):
    """
    change delimiter from $ to $$ as $ is used a lot in lammps input file
    """
    delimiter = '$$'


DEFAULT_LAMMPS_INPUT_TEMPLATE = '''\
## variables_section
$$variables_section
## post_variables_section
$$post_variables_section

## init_section
units      metal
atom_style atomic
$$init_section
## post_init_section
$$post_init_section

## read_data_section
$$read_data_section
## post_read_data_section
$$post_read_data_section

## force_field_section
$$force_field_section
## post_force_field_section
$$post_force_field_section

## md_section
$$md_section
## post_md_section
$$post_md_section

## run_section
$$run_section
## post_run_section
$$post_run_section
'''



from MDAnalysis import Universe
from MDAnalysis.lib.distances import minimize_vectors
from multiprocessing.pool import Pool
import ase.io as ai
from ase import Atom
from functools import partial

from dataclasses import dataclass
from typing import List, Tuple, NamedTuple

import fire
import numpy as np

import json
import os
import io

# TODO: use array instead of list for better performance when possible
# TODO: use numba to speed up the calculation

class AnalysisResult(NamedTuple):
    indicator_position: Tuple[float, float, float]
    transfers: List[Tuple[int, int]]


@dataclass
class SystemInfo:  # Information about the system
    initial_donor: int
    u: Universe
    cell: List[float]
    acceptor_elements: List[str]


# The relevant parameters of the algorithm
class AlgorithmParameter(NamedTuple):
    r_a: float  # The radius used to search for acceptor
    r_h: float  # The radius used to search for H
    rho_0: float  # Control the rate of the weights change
    rho_max: float  # The critical value of proton transfer
    max_depth: int  # The maximum length of the path
    g_threshold: float  # The threshold for whether to join the node to the path


class System(object):
    def __init__(self, sys_info: SystemInfo, parameter: AlgorithmParameter):
        self.u = sys_info.u
        self.cell = sys_info.cell
        self.r_a = parameter.r_a
        self.r_h = parameter.r_h
        self.g_threshold = parameter.g_threshold
        self.max_depth = parameter.max_depth
        self.rho_0 = parameter.rho_0
        self.rho_max = parameter.rho_max
        self.acceptor_elements = sys_info.acceptor_elements

    def frame_analysis(self, prev_donor: int, acceptor_query: str, time: int):
        self.u.trajectory[time]
        donor = prev_donor
        transfers = []
        list_of_paths = [[prev_donor]]
        list_of_weights = [[1]]
        for depth in range(self.max_depth):
            for j, path in enumerate(list_of_paths):
                found = False
                if depth == len(path) - 1:
                    acceptors = self.u.select_atoms(
                        f"(around {self.r_a} index {path[-1]}) and ({acceptor_query})")
                    protons = self.u.select_atoms(
                        f"(around {self.r_h} index {path[-1]}) and (name H)")
                    for i, acceptor in enumerate(acceptors.ix):
                        g, proton = self.calculate_g(
                            path[-1], acceptor, protons.ix)
                        if (g >= self.g_threshold) and (acceptor not in path):
                            found = True
                            list_of_weights.append(
                                list_of_weights[j] + [g * list_of_weights[j][-1]])
                            list_of_paths.append(path + [acceptor])
                            if proton > 0 and all(w >= 0.9 for w in list_of_weights[j]):
                                donor = acceptor
                                transfers.append((int(acceptor), int(proton)))
                if found:
                    list_of_paths.pop(j)
                    list_of_weights.pop(j)
        indicator_position = self.calculate_position(
            list_of_paths, list_of_weights)
        result = AnalysisResult(indicator_position=tuple(
            indicator_position[0]), transfers=transfers)
        return donor, result

    def calculate_g(self, donor: int, acceptor: int, protons: list):
        donor_pos = self.u.atoms[donor].position
        acceptor_pos = self.u.atoms[acceptor].position
        g_value = 0
        proton_index = -1
        for i, proton in enumerate(protons):
            proton_pos = self.u.atoms[proton].position
            r_da = minimize_vectors(acceptor_pos - donor_pos, self.cell)
            r_dh = minimize_vectors(proton_pos - donor_pos, self.cell)
            z1 = np.dot(r_dh, r_da)
            z2 = np.dot(r_da, r_da)
            z = (z1 / z2)
            p = ((self.rho_max - z) / (self.rho_max - self.rho_0))
            if p >= 1:
                g = 0
            elif p <= 0:
                g = 1
                proton_index = protons[i]
            else:
                g = -6 * (p ** 5) + 15 * (p ** 4) - 10 * (p ** 3) + 1
            g_value = g_value + g
        return g_value, proton_index

    def calculate_position(self, paths: list, weights: list):
        positions_all = []
        nodes_all = []
        weights_all = []
        for i, path in enumerate(paths):
            for j, node in enumerate(path):
                if node not in nodes_all:
                    donor_pos = self.u.atoms[path[0]].position
                    if j == 0:
                        positions_all.append(donor_pos)
                    else:
                        acceptor_pos = self.u.atoms[node].position
                        min_vector = minimize_vectors(
                            acceptor_pos - donor_pos, self.cell)
                        real_acceptor_pos = min_vector + donor_pos
                        positions_all.append(real_acceptor_pos)
                    nodes_all.append(node)
                    weights_all.append(weights[i][j])
                else:
                    index = nodes_all.index(node)
                    weights_all[index] = max(weights[i][j], weights_all[index])
        p = np.array(positions_all).reshape(-1, 3)
        w = np.array(weights_all).reshape(1, -1)
        z = w @ p
        pos_ind = z / w.sum()
        return pos_ind

    def analysis(self, initial_donor: int, out_dir: str):
        donor = initial_donor
        acceptor_query = ' or '.join(
            [f'(name {el})' for el in self.acceptor_elements])
        rand_file = io.FileIO(os.path.join(
            out_dir, f'{initial_donor}.jsonl'), 'w')
        writer = io.BufferedWriter(rand_file)
        line = (tuple(self.u.atoms[initial_donor].position.astype(float)), [])
        writer.write((json.dumps(line) + '\n').encode('utf-8'))
        for time in range(self.u.trajectory.n_frames-1):
            donor, result = self.frame_analysis(donor, acceptor_query, time+1)
            line = (result.indicator_position, result.transfers)
            writer.write((json.dumps(line) + '\n').encode('utf-8'))
        writer.flush()


def proton_transfer_detection(
        input_traj: str,
        out_dir: str,
        cell: List[float],
        acceptor_elements: List[str],
        initial_donors: List[int],
        core_num: int = 4,
        dt: float = 0.0005,
        r_a: float = 4.0,
        r_h: float = 1.3,
        rho_0: float = 1 / 2.2,
        rho_max: float = 0.5,
        max_depth: int = 4,
        g_threshold: float = 0.0001,
):
    os.makedirs(out_dir, exist_ok=True)
    u = Universe(input_traj)
    u.trajectory.ts.dt = dt
    u.dimensions = np.array(cell)

    sys_info = SystemInfo(
        initial_donor=-1,
        u=u,
        cell=cell,
        acceptor_elements=acceptor_elements)

    parameter = AlgorithmParameter(
        r_a=r_a,
        r_h=r_h,
        rho_0=rho_0,
        rho_max=rho_max,
        max_depth=max_depth,
        g_threshold=g_threshold)

    system = System(
        sys_info,
        parameter,
    )

    with Pool(processes=core_num) as pool:
        pool.map(partial(system.analysis, out_dir=out_dir), initial_donors)


def visualize_transfer(analysis_result: str, input_traj: str,output_traj: str, initial_donor: int, cell: list):
    stc_list = ai.read(input_traj, index=":")
    donor = initial_donor
    with open(os.path.join(analysis_result, f'{initial_donor}.jsonl'), mode='r') as reader:
        for i, line in enumerate(reader):
            line = json.loads(line)
            stc_list[i][donor].symbol = 'N'
            stc_list[i].set_cell(cell)
            stc_list[i].set_pbc(True)
            if line[1]:
                donor = line[1][-1][0]
            pos = line[0]
            ind = Atom('F', pos)
            stc_list[i].append(ind)
    ai.write(output_traj, stc_list)


def analysis_transfer_paths(analysis_result: str, initial_donor: int):
    donor = initial_donor
    print("transfer_paths")
    fmt = "{:^40}\t{:^8}"
    content = fmt.format("transfer_path_index", "Snapshot")
    print(f"{content}")
    with open(os.path.join(analysis_result, f'{initial_donor}.jsonl'), mode='r') as reader:
        for i, line in enumerate(reader):
            line = json.loads(line)
            if line[1]:
                for j, event in enumerate(line[1]):
                    acceptor = event[0]
                    proton = event[1]
                    content = f"{donor}({proton})->"
                    donor = acceptor
                content = content + f"{acceptor}"
                content = fmt.format(f"{content}", f"{i}")
                print(content)


def detect_type_change(analysis_result: str, atom_types: dict, donors: list):
    type_change_event = []
    type_list = []
    type_change_name = []
    for i in range(len(atom_types)):
        for j in range(i + 1):
            type_change_event.append(([], []))
            type_list.append([j, i])
            type_change_name.append(
                f"{list(atom_types.keys())[j]}<->{list(atom_types.keys())[i]}")
    for donor in donors:
        with open(os.path.join(analysis_result, f'{donor}.jsonl'), mode='r') as reader:
            change_times = 0
            change_info = []
            change = False
            for k in range(len(atom_types)):
                if donor in list(atom_types.values())[k]:
                    change_info.append((donor, k, 0))
                    change_times = 1
            for i, line in enumerate(reader):
                line = json.loads(line)
                if line[1]:
                    for j, event in enumerate(line[1]):
                        for k in range(len(atom_types)):
                            if event[0] in list(atom_types.values())[k]:
                                change_info.append((event[0], k, i))
                                change_times = change_times + 1
                                change = True
                        if change == False and change_times > 0:
                            change_info.append((event[0], -1, i))
                        if change_times == 2:
                            real_type = change_info[-1][1]
                            type = [min(change_info[0][1], change_info[-1][1]),
                                    max(change_info[0][1], change_info[-1][1])]
                            index = [x[0] for x in change_info]
                            time = [x[2] for x in change_info]
                            if index not in type_change_event[type_list.index(type)][0]:
                                type_change_event[type_list.index(
                                    type)][0].append(index)
                                type_change_event[type_list.index(
                                    type)][1].append(time)
                            change_info = [(index[-1], real_type, time[-1])]
                            change_times = 1
                            change = False
    for j, type_change in enumerate(type_change_event):
        type_change[0].sort(key=lambda x: len(x))
        type_change[1].sort(key=lambda x: len(x))
    print("proton transfer type change")
    print("-------------------------------------")
    fmt = "{:^25}\t{:^15}\t{:^15}"
    content = fmt.format("Path_index", "start_Snapshot", "end_Snapshot")
    print(f"{content}")
    for i in range(len(type_change_name)):
        print(type_change_name[i])
        for j in range(len(type_change_event[i][0])):
            content = ' -> '.join([f'{el}' for el in type_change_event[i][0][j]])
            content = fmt.format(f"{content}", f"{type_change_event[i][1][j][0]}",
                                 f"{type_change_event[i][1][j][-1]}")
            print(f"{content}")


if __name__ == '__main__':
    fire.Fire({
        "analyze": proton_transfer_detection,
        "visualize": visualize_transfer,
        "show-transfer-paths": analysis_transfer_paths,
        "show-type-change": detect_type_change
    })



from abc import abstractmethod, ABC
from ai2_kit.core.artifact import Artifact, ArtifactMap
from ai2_kit.core.resource_manager import ResourceManager
from typing import List
from dataclasses import dataclass


@dataclass
class BaseCllContext:
    path_prefix: str
    resource_manager: ResourceManager


class ICllLabelOutput(ABC):
    @abstractmethod
    def get_labeled_system_dataset(self) -> List[Artifact]:
        ...

class ICllTrainOutput(ABC):
    @abstractmethod
    def get_mlp_models(self) -> List[Artifact]:
        ...

    @abstractmethod
    def get_training_dataset(self) -> List[Artifact]:
        ...

class ICllExploreOutput(ABC):
    @abstractmethod
    def get_model_devi_dataset(self) -> List[Artifact]:
        ...

class ICllSelectorOutput(ABC):
    @abstractmethod
    def get_model_devi_dataset(self) -> List[Artifact]:
        ...

    @abstractmethod
    def get_passing_rate(self) -> float:
        ...


def init_artifacts(artifacts: ArtifactMap):
    for key, artifact in artifacts.items():
        artifact.attrs.setdefault('ancestor', key)



from ruamel.yaml import YAML
from pathlib import Path
from typing import Tuple, List, TypeVar
from dataclasses import field

import shortuuid
import hashlib
import base64
import copy
import os

from .log import get_logger

logger = get_logger(__name__)

EMPTY = object()

def default_mutable_field(obj):
    return field(default_factory=lambda: copy.copy(obj))


def load_yaml_file(path: Path):
    yaml = YAML(typ='safe')
    JoinTag.register(yaml)
    ReadTag.register(yaml)
    return yaml.load(path)


def load_yaml_files(*paths: Tuple[Path]):
    d = {}
    for path in paths:
        print('load yaml file: ', path)
        d = merge_dict(d, load_yaml_file(Path(path)))  # type: ignore
    return d


def s_uuid():
    """short uuid"""
    return shortuuid.uuid()


def sort_unique_str_list(l: List[str]) -> List[str]:
    """remove duplicate str and sort"""
    return list(sorted(set(l)))


T = TypeVar('T')
def flatten(l: List[List[T]]) -> List[T]:
    return [item for sublist in l for item in sublist]


def format_env_string(s: str) -> str:
    return s.format(**os.environ)


def split_list(l: List[T], n: int) -> List[List[T]]:
    """split list into n chunks"""
    # ref: https://stackoverflow.com/questions/2130016/splitting-a-list-into-n-parts-of-approximately-equal-length
    k, m = divmod(len(l), n)
    return [l[i*k+min(i, m) : (i+1)*k+min(i+1, m)] for i in range(n)]


def short_hash(s: str) -> str:
    """short hash string"""
    digest = hashlib.sha1(s.encode('utf-8')).digest()
    # use urlsafe encode to avoid '/' in the string, as it will cause problem in file path
    return base64.urlsafe_b64encode(digest).decode('utf-8')[:-2]

async def to_awaitable(value: T) -> T:
    return value


class JoinTag:
    """a tag to join strings in a list"""

    yaml_tag = u'!join'

    @classmethod
    def from_yaml(cls, constructor, node):
        seq = constructor.construct_sequence(node)
        return ''.join([str(i) for i in seq])

    @classmethod
    def to_yaml(cls, dumper, data):
        # do nothing
        return dumper.represent_sequence(cls.yaml_tag, data)

    @classmethod
    def register(cls, yaml: YAML):
        yaml.register_class(cls)


class ReadTag:
    """a tag to read string from file"""

    yaml_tag = u'!read'

    @classmethod
    def from_yaml(cls, constructor, node):
        seq = constructor.construct_sequence(node)
        path = os.path.join(*seq)
        with open(path, 'r') as f:
            return f.read()

    @classmethod
    def to_yaml(cls, dumper, data):
        # do nothing
        return dumper.represent_sequence(cls.yaml_tag, data)

    @classmethod
    def register(cls, yaml: YAML):
        yaml.register_class(cls)


def __export_remote_functions():
    """cloudpickle compatible: https://stackoverflow.com/questions/75292769"""

    def merge_dict(lo: dict, ro: dict, path=None):
        """
        Merge two dict, the left dict will be overridden.
        Note: list will be replaced instead of merged.
        """
        if path is None:
            path = []
        for key, value in ro.items():
            if key in lo:
                current_path = path + [ str(key) ]
                if isinstance(lo[key], dict) and isinstance(value, dict):
                    merge_dict(lo[key], value, current_path)
                else:
                    print('.'.join(current_path) + ' has been overridden')
                    lo[key] = value
            else:
                lo[key] = value
        return lo

    def dict_nested_get(d: dict, keys: List[str], default=EMPTY):
        """get value from nested dict"""
        for key in keys:
            if key not in d and default is not EMPTY:
                return default
            d = d[key]
        return d

    def dict_nested_set(d: dict, keys: List[str], value):
        """set value to nested dict"""
        for key in keys[:-1]:
            d = d[key]
        d[keys[-1]] = value
    return merge_dict, dict_nested_get, dict_nested_set

(
    merge_dict,
    dict_nested_get,
    dict_nested_set,
) = __export_remote_functions()


import ase.io
from ase import Atoms
from typing import List

class AseHelper:
    def __init__(self):
        self._atoms_list: List[Atoms] = []

    def read(self, filename: str, **kwargs):
        kwargs.setdefault('index', ':')
        data = ase.io.read(filename, **kwargs)
        if not isinstance(data, list):
            data = [data]
        self._atoms_list = data
        return self

    def write(self, filename: str, **kwargs):
        ase.io.write(filename, self._atoms_list, **kwargs)

    def set_cell(self, cell, scale_atoms=False, apply_constraint=True):
        for atoms in self._atoms_list:
            atoms.set_cell(cell, scale_atoms=scale_atoms, apply_constraint=apply_constraint)
        return self

    def set_pbc(self, pbc):
        print(pbc)
        for atoms in self._atoms_list:
            atoms.set_pbc(pbc)
        return self

    def write_each_frame(self, filename: str, **kwargs):
        for i, atoms in enumerate(self._atoms_list):
            ase.io.write(filename.format(i=i), atoms, **kwargs)



from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.script import BashScript, BashStep, BashTemplate
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.util import merge_dict, dict_nested_get, split_list
from ai2_kit.core.log import get_logger

from typing import List, Union, Tuple
from pydantic import BaseModel
from dataclasses import dataclass

import copy
import os
import io

from .data_helper import LammpsOutputHelper, XyzHelper, Cp2kOutputHelper, ase_atoms_to_cp2k_input_data
from .cll import ICllLabelOutput, BaseCllContext
from .util import loads_cp2k_input, load_cp2k_input, dump_cp2k_input

logger = get_logger(__name__)

class GenericCp2kInputConfig(BaseModel):
    init_system_files: List[str] = []
    limit: int = 50
    input_template: Union[dict, str]
    """
    Input template for cp2k. Could be a dict or content of a cp2k input file.

    Note:
    If you are using files in input templates, it is recommended to use artifact name instead of literal path.
    String starts with '@' will be treated as artifact name.
    For examples, FORCE_EVAL/DFT/BASIS_SET_FILE_NAME = @cp2k/basic_set.
    You can still use literal path, but it is not recommended.
    """

class GenericCp2kContextConfig(BaseModel):
    script_template: BashTemplate
    cp2k_cmd: str = 'cp2k'
    concurrency: int = 5

@dataclass
class GenericCp2kInput:
    config: GenericCp2kInputConfig
    system_files: List[Artifact]
    type_map: List[str]
    initiated: bool = False


@dataclass
class GenericCp2kContext(BaseCllContext):
    config: GenericCp2kContextConfig


@dataclass
class GenericCp2kOutput(ICllLabelOutput):
    cp2k_outputs: List[Artifact]

    def get_labeled_system_dataset(self):
        return self.cp2k_outputs


async def generic_cp2k(input: GenericCp2kInput, ctx: GenericCp2kContext) -> GenericCp2kOutput:
    executor = ctx.resource_manager.default_executor

    # For the first round
    if not input.initiated:
        input.system_files += ctx.resource_manager.get_artifacts(input.config.init_system_files)

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [tasks_dir] = executor.setup_workspace(work_dir, ['tasks'])

    # prepare input template
    if isinstance(input.config.input_template, str):
        input_template = loads_cp2k_input(input.config.input_template)
    else:
        input_template = copy.deepcopy(input.config.input_template)

    # resolve data files
    lammps_dump_files: List[Artifact] = []
    xyz_files: List[Artifact] = []

    # TODO: support POSCAR in the future
    # TODO: refactor the way of handling different file formats
    system_files = ctx.resource_manager.resolve_artifacts(input.system_files)
    for system_file in system_files:
        if LammpsOutputHelper.is_match(system_file):
            lammps_out = LammpsOutputHelper(system_file)
            lammps_dump_files.extend(lammps_out.get_passed_dump_files())
        elif XyzHelper.is_match(system_file):
            xyz_files.append(system_file)
        else:
            raise ValueError(f'unsupported format {system_file.url}: {system_file.format}')

    # create task dirs and prepare input files
    cp2k_task_dirs = []
    if lammps_dump_files or xyz_files:
        cp2k_task_dirs = executor.run_python_fn(make_cp2k_task_dirs)(
            lammps_dump_files=[a.to_dict() for a in lammps_dump_files],
            xyz_files=[a.to_dict() for a in xyz_files],
            type_map=input.type_map,
            base_dir=tasks_dir,
            input_template=input_template,
            limit=input.config.limit,
        )
    else:
        logger.warn('no available candidates, skip')
        return GenericCp2kOutput(cp2k_outputs=[])

    # build commands
    steps = []
    for cp2k_task_dir in cp2k_task_dirs:
        steps.append(BashStep(
            cwd=cp2k_task_dir['url'],
            cmd=[ctx.config.cp2k_cmd, '-i input.inp 1>> output 2>> output'],
            checkpoint='cp2k',
        ))

    # submit tasks and wait for completion
    jobs = []
    for i, steps_group in enumerate(split_list(steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps_group,
        )
        job = executor.submit(script.render(), cwd=tasks_dir,
                              checkpoint_key=f'submit-job/cp2k/{i}:{tasks_dir}')
        jobs.append(job)
    jobs = await gather_jobs(jobs, max_tries=2)

    cp2k_outputs = [Artifact.of(
        url=a['url'],
        format=Cp2kOutputHelper.format,
        executor=executor.name,
        attrs=a['attrs'],
    ) for a in cp2k_task_dirs]

    return GenericCp2kOutput(cp2k_outputs=cp2k_outputs)


def __export_remote_functions():

    def make_cp2k_task_dirs(lammps_dump_files: List[ArtifactDict],
                            xyz_files: List[ArtifactDict],
                            type_map: List[str],
                            input_template: dict,
                            base_dir: str,
                            limit: int = 0,
                            input_file_name: str = 'input.inp',
                            ) -> List[ArtifactDict]:
        """Generate CP2K input files from LAMMPS dump files or XYZ files."""
        import ase.io
        from ase import Atoms

        task_dirs = []
        atoms_list: List[Tuple[ArtifactDict, Atoms]] = []

        # read atoms
        for dump_file in lammps_dump_files:
            atoms_list += [
                (dump_file, atoms)
                for atoms in ase.io.read(dump_file['url'], ':', format='lammps-dump-text', order=False, specorder=type_map)
            ]  # type: ignore
        for xyz_file in xyz_files:
            atoms_list += [
                (xyz_file, atoms)
                for atoms in ase.io.read(xyz_file['url'], ':', format='extxyz')
            ]  # type: ignore

        if limit > 0:
            atoms_list = atoms_list[:limit]

        for i, (file, atoms) in enumerate(atoms_list):
            # create task dir
            task_dir = os.path.join(base_dir, f'{str(i).zfill(6)}')
            os.makedirs(task_dir, exist_ok=True)

            # TODO: should also support input_template
            # find input template in data_file attrs, if not found, use input_template as default
            input_data_file = dict_nested_get(file, ['attrs', 'cp2k', 'input_template_file'],  None)  # type: ignore
            if isinstance(input_data_file, str):
                with open(input_data_file, 'r') as f:
                    input_data = load_cp2k_input(f)
            else:
                input_data = copy.deepcopy(input_template)

            coords, cell = ase_atoms_to_cp2k_input_data(atoms)
            merge_dict(input_data, {
                'FORCE_EVAL': {
                    'SUBSYS': {
                        # FIXME: this is a dirty hack, we should make dump_cp2k_input support COORD
                        'COORD': dict.fromkeys(coords, ''),
                        'CELL': {
                            'A': ' '.join(map(str, cell[0])),
                            'B': ' '.join(map(str, cell[1])),
                            'C': ' '.join(map(str, cell[2])),
                        }
                    }
                }
            })
            with open(os.path.join(task_dir, input_file_name), 'w') as f:
                dump_cp2k_input(input_data, f)

            task_dirs.append({
                'url': task_dir,
                'attrs': file['attrs'],
            })

        return task_dirs

    return make_cp2k_task_dirs

make_cp2k_task_dirs = __export_remote_functions()



from ai2_kit.core.artifact import Artifact
from ai2_kit.core.log import get_logger

from typing import List
from io import StringIO
from pydantic import BaseModel
from dataclasses import dataclass
import pandas as pd

from .data_helper import LammpsOutputHelper
from .cll import ICllSelectorOutput, BaseCllContext

logger = get_logger(__name__)

class ThresholdSelectorInputConfig(BaseModel):
    f_trust_lo: float
    f_trust_hi: float


@dataclass
class ThresholdSelectorContext(BaseCllContext):
    ...

@dataclass
class ThresholdSelectorOutput(ICllSelectorOutput):
    model_devi_data: List[Artifact]
    passing_rate: float

    def get_model_devi_dataset(self):
        return self.model_devi_data

    def get_passing_rate(self) -> float:
        return self.passing_rate

@dataclass
class ThresholdSelectorInput:
    config: ThresholdSelectorInputConfig
    model_devi_data: List[Artifact]
    model_devi_out_filename: str

    def set_model_devi_dataset(self, data: List[Artifact]):
        self.model_devi_data = data

async def threshold_selector(input: ThresholdSelectorInput, ctx: ThresholdSelectorContext):
    executor = ctx.resource_manager.default_executor

    f_trust_lo = input.config.f_trust_lo
    f_trust_hi = input.config.f_trust_hi
    col_force = 'avg_devi_f'
    logger.info('criteria: %f <= %s < %f ', f_trust_lo, col_force, f_trust_hi)

    total_count = 0
    passed_count = 0

    # TODO: support output of different software
    for candidate in input.model_devi_data:
        if LammpsOutputHelper.is_match(candidate):
            model_devi_out_file = LammpsOutputHelper(candidate).get_model_devi_file(input.model_devi_out_filename).url
        else:
            raise ValueError('unknown model_devi_data types')

        logger.info('start to analysis file: %s', model_devi_out_file)
        text = executor.load_text(model_devi_out_file)

        df = pd.read_csv(StringIO(text.lstrip('#')), delim_whitespace=True)
        # layout:
        #        step  max_devi_v  min_devi_v  avg_devi_v  max_devi_f  min_devi_f  avg_devi_f
        # 0        0    0.006793    0.000672    0.003490    0.143317    0.005612    0.026106
        # 1      100    0.006987    0.000550    0.003952    0.128178    0.006042    0.022608

        passed_df   = df[df[col_force] < f_trust_lo]
        selected_df = df[(df[col_force] >= f_trust_lo) & (df[col_force] < f_trust_hi)]
        rejected_df = df[df[col_force] >= f_trust_hi]

        logger.info('result: total: %d, passed: %d, selected: %d, rejected: %d', len(df), len(passed_df), len(selected_df), len(rejected_df))

        candidate.attrs['all'] = df.step.tolist()
        candidate.attrs['passed']   = passed_df.step.tolist()
        candidate.attrs['selected'] = selected_df.step.tolist()
        candidate.attrs['rejected'] = rejected_df.step.tolist()

        total_count += len(df)
        passed_count += len(passed_df)

    return ThresholdSelectorOutput(
        model_devi_data=input.model_devi_data,
        passing_rate=passed_count / total_count,
    )



from ai2_kit.core.artifact import Artifact, ArtifactDict

from typing import List, Tuple, Optional
from ase import Atoms
import os
import re

from .constant import MODEL_DEVI_OUT, LAMMPS_TRAJ_DIR, LAMMPS_TRAJ_SUFFIX


class DataHelper:

    format: Optional[str] = None
    suffix: Optional[str] = None
    pattern: Optional[re.Pattern] = None

    @classmethod
    def is_match(cls, artifact: Artifact) -> bool:
        if cls.format and artifact.format == cls.format:
            return True
        if cls.suffix and artifact.url.endswith(cls.suffix):
            return True
        file_name = os.path.basename(artifact.url)
        if cls.pattern and cls.pattern.match(file_name):
            return True
        return False

    def __init__(self, artifact: Artifact) -> None:
        self.artifact = artifact


class LammpsOutputHelper(DataHelper):
    format = 'lammps/output-dir'

    def get_model_devi_file(self, filename: str) -> Artifact:
        return self.artifact.join(filename)

    def get_passed_dump_files(self) -> List[Artifact]:
        return [self.artifact.join(LAMMPS_TRAJ_DIR, f'{i}{LAMMPS_TRAJ_SUFFIX}') for i in self.artifact.attrs['passed']]

class PoscarHelper(DataHelper):
    format = 'vasp/poscar'
    pattern = re.compile(r'POSCAR')


class XyzHelper(DataHelper):
    format = 'extxyz'
    suffix = '.xyz'

class DeepmdNpyHelper(DataHelper):
    format = 'deepmd/npy'

class DeepmdModelHelper(DataHelper):
    format = 'deepmd/model'

class Cp2kOutputHelper(DataHelper):
    format = 'cp2k-output-dir'

class VaspOutputHelper(DataHelper):
    format = 'vasp-output-dir'



def __ase_atoms_to_cp2k_input_data():
    """workaround for cloudpickle issue"""
    def ase_atoms_to_cp2k_input_data(atoms: Atoms) -> Tuple[List[str], List[List[float]]]:
        coords = [atom.symbol + ' ' + ' '.join(str(x) for x in atom.position) for atom in atoms] # type: ignore
        cell = [list(row) for row in atoms.cell]  # type: ignore
        return (coords, cell)
    return ase_atoms_to_cp2k_input_data
ase_atoms_to_cp2k_input_data = __ase_atoms_to_cp2k_input_data()


def __convert_to_deepmd_npy():
    """workaround for cloudpickle issue"""
    def covert_to_deepmd_npy(cp2k_outputs: List[ArtifactDict], base_dir: str, type_map: List[str]):
        import dpdata
        from itertools import groupby

        atoms_list: List[Tuple[ArtifactDict, Atoms]] = []
        for cp2k_output in cp2k_outputs:
            dp_system = dpdata.LabeledSystem(os.path.join(cp2k_output['url'], 'output'), fmt='cp2k/output', type_map=type_map)
            atoms_list += [
                (cp2k_output, atoms)
                for atoms in dp_system.to_ase_structure()  # type: ignore
            ]

        output_dirs = []
        # group dataset by ancestor key
        for i, (key, atoms_group) in enumerate(groupby(atoms_list, key=lambda x: x[0]['attrs']['ancestor'])):
            output_dir = os.path.join(base_dir, key.replace('/', '_'))
            atoms_group = list(item[1] for item in atoms_group)
            if not atoms_group:
                continue
            dp_system = dpdata.LabeledSystem(atoms_group[0], fmt='ase/structure')
            for atoms in atoms_group[1:]:
                dp_system += dpdata.LabeledSystem(atoms, fmt='ase/structure')
            dp_system.to_deepmd_npy(output_dir, set_size=len(dp_system))  # type: ignore
            # TODO: return ArtifactDict
            output_dirs.append(output_dir)
        return output_dirs

    return covert_to_deepmd_npy
convert_to_deepmd_npy = __convert_to_deepmd_npy()


def __convert_to_lammps_input_data():
    """workaround for cloudpickle issue"""
    def convert_to_lammps_input_data(poscar_files: List[ArtifactDict], base_dir: str, type_map: List[str]):
        import dpdata
        import os
        lammps_data_files = []
        for i, poscar_file in enumerate(poscar_files):
            output_file = os.path.join(base_dir, f'{i:06d}.lammps.data')
            dpdata.System(poscar_file['url'], fmt='vasp/poscar', type_map=type_map).to_lammps_lmp(output_file)  # type: ignore
            lammps_data_files.append({
                'url': output_file,
                'attrs': poscar_file['attrs'],
            })
        return lammps_data_files

    return convert_to_lammps_input_data
convert_to_lammps_input_data = __convert_to_lammps_input_data()



from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.script import BashScript, BashStep, BashTemplate
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.util import merge_dict, dict_nested_get, dict_nested_set, split_list
from ai2_kit.core.log import get_logger

from typing import List, Union
from pydantic import BaseModel
from dataclasses import dataclass

from typing import List, Union, Optional, Tuple
from pymatgen.io.vasp.inputs import Incar, Kpoints
from pydantic import BaseModel
from dataclasses import dataclass

import copy
import os

from .data_helper import LammpsOutputHelper, XyzHelper, VaspOutputHelper
from .cll import ICllLabelOutput, BaseCllContext

logger = get_logger(__name__)

class GenericVaspInputConfig(BaseModel):
    init_system_files: List[str] = []
    limit: int = 50
    input_template: Union[dict, str]
    potcar_source: Union[dict, list]
    kpoints_template: Optional[str] = None
    """
    Input template for VASP. Could be a dict or content of a VASP input file.

    Note:
    If you are using files in input templates, it is recommended to use artifact name instead of literal path.
    String starts with '@' will be treated as artifact name.
    For examples, input_template = @vasp/INCAR.
    You can still use literal path, but it is not recommended.
    """

class GenericVaspContextConfig(BaseModel):
    script_template: BashTemplate
    vasp_cmd: str = 'vasp_std'
    concurrency: int = 5

@dataclass
class GenericVaspInput:
    config: GenericVaspInputConfig
    system_files: List[Artifact]
    type_map: List[str]
    initiated: bool = False


@dataclass
class GenericVaspContext(BaseCllContext):
    config: GenericVaspContextConfig


@dataclass
class GenericVaspOutput(ICllLabelOutput):
    vasp_outputs: List[Artifact]

    def get_labeled_system_dataset(self):
        return self.vasp_outputs


async def generic_vasp(input: GenericVaspInput, ctx: GenericVaspContext) -> GenericVaspOutput:
    executor = ctx.resource_manager.default_executor

    # For the first round
    if not input.initiated:
        input.system_files += ctx.resource_manager.get_artifacts(input.config.init_system_files)

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [tasks_dir] = executor.setup_workspace(work_dir, ['tasks'])

    # prepare input template
    if isinstance(input.config.input_template, str):
        input_template = input.config.input_template
        if input_template.startswith('@'):
            input_template = \
                ctx.resource_manager.resolve_artifact(input_template[1:])[0].url
        input_template = Incar.from_file(input_template).as_dict()
    else:
        input_template = copy.deepcopy(input.config.input_template)

    # prepare potcar
    if isinstance(input.config.potcar_source, dict):
        potcar_source = input.config.potcar_source
    else:
        # default: use same sequence as type_map
        if len(input.config.potcar_source) >= len(input.type_map):
            potcar_source = {
                k: v for k, v in zip(input.type_map, input.config.potcar_source)
            }
        else:
            raise ValueError('potcar_source should not be shorter than type_map')
        
    for k, v in potcar_source.items():
        if v.startswith('@'):
            potcar_source[k] = \
                ctx.resource_manager.resolve_artifact(v[1:])[0].url

    # prepare kpoints
    kpoints_template = input.config.kpoints_template
    if kpoints_template:
        if kpoints_template.startswith('@'):
            logger.info(f'resolve artifact {kpoints_template}')
            kpoints_template = \
                ctx.resource_manager.resolve_artifact(kpoints_template[1:])[0].url
        kpoints_template = Kpoints.from_file(kpoints_template).as_dict()
    else:
        kpoints_template = None

    # resolve data files
    lammps_dump_files: List[Artifact] = []
    xyz_files: List[Artifact] = []

    # TODO: support POSCAR in the future
    # TODO: refactor the way of handling different file formats
    system_files = ctx.resource_manager.resolve_artifacts(input.system_files)
    for system_file in system_files:
        if LammpsOutputHelper.is_match(system_file):
            lammps_out = LammpsOutputHelper(system_file)
            lammps_dump_files.extend(lammps_out.get_passed_dump_files())
        elif XyzHelper.is_match(system_file):
            xyz_files.append(system_file)
        else:
            raise ValueError(f'unsupported format {system_file.url}: {system_file.format}')

    # create task dirs and prepare input files
    vasp_task_dirs = []
    if lammps_dump_files or xyz_files:
        vasp_task_dirs = executor.run_python_fn(make_vasp_task_dirs)(
            lammps_dump_files=[a.to_dict() for a in lammps_dump_files],
            xyz_files=[a.to_dict() for a in xyz_files],
            type_map=input.type_map,
            base_dir=tasks_dir,
            input_template=input_template,
            potcar_source=potcar_source,
            kpoints_template=kpoints_template,
            limit=input.config.limit,
        )
    else:
        logger.warn('no available candidates, skip')
        return GenericVaspOutput(vasp_outputs=[])

    # build commands
    steps = []
    for vasp_task_dir in vasp_task_dirs:
        steps.append(BashStep(
            cwd=vasp_task_dir['url'],
            cmd=[ctx.config.vasp_cmd, ' 1>> output 2>> output'],
            checkpoint='vasp',
        ))

    # run tasks
    jobs = []
    for i, steps_group in enumerate(split_list(steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps,
        )
        job = executor.submit(script.render(), cwd=tasks_dir,
                              checkpoint_key=f'submit-job/vasp/{i}:{tasks_dir}')
        jobs.append(job)
    jobs = await gather_jobs(jobs, max_tries=2)

    vasp_outputs = [Artifact.of(
        url=a['url'],
        format=VaspOutputHelper.format,
        executor=executor.name,
        attrs=a['attrs'],
    ) for a in vasp_task_dirs]

    return GenericVaspOutput(vasp_outputs=vasp_outputs)


def __make_vasp_task_dirs():
    def make_vasp_task_dirs(lammps_dump_files: List[ArtifactDict],
                            xyz_files: List[ArtifactDict],
                            type_map: List[str],
                            input_template: dict,
                            potcar_source: dict,
                            base_dir: str,
                            kpoints_template: Optional[dict] = None,
                            limit: int = 0
                            ) -> List[ArtifactDict]:
        """Generate VASP input files from LAMMPS dump files or XYZ files."""

        import ase.io
        from ase import Atoms
        from ase.io.vasp import _symbol_count_from_symbols

        task_dirs = []
        atoms_list: List[Tuple[ArtifactDict, Atoms]] = []

        # read atoms
        for dump_file in lammps_dump_files:
            atoms_list += [
                (dump_file, atoms)
                for atoms in ase.io.read(dump_file['url'], ':', format='lammps-dump-text', order=False, specorder=type_map)
            ]  # type: ignore
        for xyz_file in xyz_files:
            atoms_list += [
                (xyz_file, atoms)
                for atoms in ase.io.read(xyz_file['url'], ':', format='extxyz')
            ]  # type: ignore

        if limit > 0:
            atoms_list = atoms_list[:limit]

        for i, (file, atoms) in enumerate(atoms_list):
            # create task dir
            task_dir = os.path.join(base_dir, f'{str(i).zfill(6)}')
            os.makedirs(task_dir, exist_ok=True)

            # create input file
            input_data = copy.deepcopy(input_template)
            input_data = dict_nested_get(
                file, ['attrs', 'vasp', 'input_data'], input_data # type: ignore
            )
            incar = Incar.from_dict(input_data)
            incar.write_file(os.path.join(task_dir, 'INCAR'))

            # create POSCAR
            elements_all = atoms.get_chemical_symbols()
            elements = [
                item[0] for item in _symbol_count_from_symbols(elements_all)
            ]
            ase.io.write(
                os.path.join(task_dir, 'POSCAR'), atoms, format='vasp5'
            )

            # create POTCAR
            with open(os.path.join(task_dir, 'POTCAR'), 'w') as out_file:
                for element in elements:
                    with open(potcar_source[element], 'r') as in_file:
                        out_file.write(in_file.read())

            # create KPOINTS
            kpoints_template = dict_nested_get(
                file, ['attrs', 'vasp', 'kpoints_template'] # type: ignore
            )
            if kpoints_template:
                kpoints = Kpoints.from_dict(kpoints_template)
                kpoints.write_file(os.path.join(task_dir, 'KPOINTS'))

            # inherit attrs from input file
            # TODO: inherit only ancestor key should be enough
            task_dirs.append({
                'url': task_dir,
                'attrs': file['attrs'],
            })

        return task_dirs

    return make_vasp_task_dirs
make_vasp_task_dirs = __make_vasp_task_dirs()



from typing import TypeVar, Union, Callable, NamedTuple, Optional
from threading import Lock
import functools
import cloudpickle
import os
import inspect

from .log import get_logger
from .util import to_awaitable

logger = get_logger(__name__)

_lock = Lock()
_checkpoint_file: Optional[str] = None
_checkpoint_data: Optional[dict] = None


class FnInfo(NamedTuple):
    fn_name: str
    args: tuple
    kwargs: dict
    call_site: str


KeyFn = Callable[[FnInfo], str]

EMPTY = object()


def set_checkpoint_file(path: str):
    global _checkpoint_file
    if _checkpoint_file is not None:
        raise RuntimeError(
            "checkpoint path has been set to {}".format(_checkpoint_file))
    _checkpoint_file = path
    _load_checkpoint()


def apply_checkpoint(key_fn: Union[str, KeyFn], disable = False):
    """
    apply checkpoint for function.

    Note: This checkpoint implementation doesn't support multiprocess.
    To support multiple process we need to have a dedicated background process to read/write checkpoint,
    which will require message queue (e.g. nanomsg or nng) to implement it.

    Example:

    >>> set_checkpoint_file('/tmp/test.ckpt')
    >>> task_fn = lambda a, b: a + b
    >>> checkpoint('task_1+2')(task_fn)(1, 2)
    """

    call_site = inspect.getframeinfo(inspect.stack()[1][0])

    T = TypeVar('T', bound=Callable)
    def _checkpoint(fn: T) -> T:

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            fn_info = FnInfo(
                fn_name=fn.__name__,
                args=args,
                kwargs=kwargs,
                call_site=f'{call_site.filename}:{call_site.lineno}',
            )

            key = key_fn if isinstance(key_fn, str) else key_fn(fn_info)

            if disable or _checkpoint_file is None:
                return fn(*args, **kwargs)

            ret = _get_checkpoint(key)
            if ret is not EMPTY:
                return ret

            ret = fn(*args, **kwargs)
            if inspect.isawaitable(ret):
                async def _wrap_fn():
                    _ret = await ret
                    _set_checkpoint(key, _ret, fn_info, True)
                    return _ret
                return _wrap_fn()
            else:
                _set_checkpoint(key, ret, fn_info, False)
                return ret

        return wrapper # type: ignore
    return _checkpoint


def _load_checkpoint():
    global _checkpoint_data
    if _checkpoint_data is not None:
        return
    assert _checkpoint_file is not None, '_checkpoint_path should not be None!'
    if os.path.exists(_checkpoint_file):
        with open(_checkpoint_file, 'rb') as f:
            _checkpoint_data = cloudpickle.load(f)
    else:
        _checkpoint_data = dict()


def _dump_checkpoint():
    assert _checkpoint_data is not None, '_checkpoint_data should not be None!'
    with open(_checkpoint_file, 'wb') as f:  # type: ignore
        cloudpickle.dump(_checkpoint_data, f)


def _get_checkpoint(key: str):
    try:
        with _lock:
            _load_checkpoint()
            assert _checkpoint_data is not None
            value = _checkpoint_data.get(key, None)
            if value is None:
                return EMPTY
            logger.info(f"Hit checkpoint: {key}")
            if value['is_awaitable']:
                return to_awaitable(value['return'])
            else:
                return value['return']

    except Exception as e:
        logger.error(f"Fail to get checkpoint: {key}", e)
        return EMPTY


def _set_checkpoint(key: str, value, info: FnInfo, is_awaitable: bool = False):
    try:
        with _lock:
            assert _checkpoint_data is not None
            # args, kwargs may contain unpickable objects
            _checkpoint_data[key] = {
                'return': value,
                'is_awaitable': is_awaitable,
                'info': {
                    'fn_name': info.fn_name,
                    'call_site': info.call_site,
                }
            }
            _dump_checkpoint()
    except Exception as e:
        logger.error('Fail to set checkpoint', e)


def del_checkpoint(key: str):
    try:
        with _lock:
            _load_checkpoint()
            assert _checkpoint_data is not None
            if key in _checkpoint_data:
                del _checkpoint_data[key]
                _dump_checkpoint()
    except Exception as e:
        logger.error('Fail to delete checkpoint', e)


class CheckpointCmd:
    """checkpoint command line interface"""
    def __init__(self) -> None:
        ...

    def load(self, file):
        set_checkpoint_file(file)
        return self

    def ls(self):
        '''list all the checkpoint entries in the checkpoint file'''
        assert _checkpoint_data is not None
        for i, (key, value) in enumerate(_checkpoint_data.items()):
            print('\n'.join([
                '=' * 80,
                f'Key:        \t{key}',
                f'Call Site:  \t{value["info"]["call_site"]}',
                f'Function:   \t{value["info"]["fn_name"]}',
            ]))
        print('=' * 80)

    def rm(self, prefix: str):
        """remove checkpoint entries with the given prefix"""
        assert _checkpoint_data is not None

        for key in list(_checkpoint_data.keys()):
            if key.startswith(prefix):
                logger.info(f"Remove checkpoint: {key}")
                del _checkpoint_data[key]
        _dump_checkpoint()



from pydantic import BaseModel
from typing import Optional, List, Union
import shlex


def exit_on_error_statment(v='__EXITCODE__'):
    return f"{v}=$?; if [ ${v} -ne 0 ]; then exit ${v}; fi"


def eoe_step(cmd: str):
    """
    exit on error step
    """
    return '\n'.join([cmd, exit_on_error_statment()])


class BashTemplate(BaseModel):
    shebang: str = '#!/bin/bash'
    header: str = ''
    setup: str = ''
    teardown: str = ''


class BashStep(BaseModel):
    cmd: Union[str, List[str]]
    cwd: Optional[str]
    checkpoint: Optional[str]
    exit_on_error: bool = True

    def render(self):
        if isinstance(self.cmd, list):
            self.cmd = ' '.join(self.cmd)

        rendered_step = '\n'.join([
            '#' * 80,
            self.cmd,
            '#' * 80,
        ])

        if self.exit_on_error:
            rendered_step = '\n'.join([
                rendered_step,
                exit_on_error_statment()
            ])

        if self.checkpoint:
            checkpoint = shlex.quote(self.checkpoint + '.checkpoint')
            msg = shlex.quote(f"pass {checkpoint}")

            rendered_step = '\n'.join([
                f'if [ -f {checkpoint} ]; then echo {msg}; else',
                rendered_step,
                f'touch {checkpoint}; fi  # create checkpoint on success',
            ])

        if self.cwd:
            cwd = shlex.quote(self.cwd)
            rendered_step = '\n'.join([
                f'pushd {cwd} || exit 1',
                rendered_step,
                'popd',
            ])

        return rendered_step


BashSteps = List[Union[str, BashStep]]


class BashScript(BaseModel):
    template: Optional[BashTemplate]
    steps: BashSteps

    def render(self):
        if self.template is None:
            return _render_bash_steps(self.steps)

        return '\n'.join([
            self.template.shebang,
            self.template.header,
            self.template.setup,
            _render_bash_steps(self.steps),
            self.template.teardown,
        ])


def _render_bash_steps(steps: BashSteps):
    rendered_steps = []

    for step in steps:
        if isinstance(step, str):
            rendered_steps.append(step)
        else:
            assert isinstance(step, BashStep)
            rendered_steps.append(step.render())
    return '\n\n'.join(rendered_steps)



from typing import List, Callable, Optional, Awaitable
from enum import Enum
from abc import abstractmethod
import time
import asyncio

from .future import IFuture

# Copy from parsl
class JobState(bytes, Enum):
    """Defines a set of states that a job can be in"""
    def __new__(cls, value: int, terminal: bool, status_name: str) -> "JobState":
        obj = bytes.__new__(cls, [value])
        obj._value_ = value
        obj.terminal = terminal
        obj.status_name = status_name
        return obj

    value: int
    terminal: bool
    status_name: str

    UNKNOWN = (0, False, "UNKNOWN")
    PENDING = (1, False, "PENDING")
    RUNNING = (2, False, "RUNNING")
    CANCELLED = (3, True, "CANCELLED")
    COMPLETED = (4, True, "COMPLETED")
    FAILED = (5, True, "FAILED")
    TIMEOUT = (6, True, "TIMEOUT")
    HELD = (7, False, "HELD")

class TimeoutError(RuntimeError):
    ...

class JobFuture(IFuture[JobState]):

    @abstractmethod
    def get_job_state(self) -> JobState:
        ...

    @abstractmethod
    def cancel(self):
        ...

    @abstractmethod
    def is_success(self) -> bool:
        ...

    @abstractmethod
    def resubmit(self) -> 'JobFuture':
        ...

async def gather_jobs(jobs: List[JobFuture], timeout = float('inf'), max_tries: int = 1, raise_error=True) -> List[JobState]:
    async def wait_job(job: JobFuture) -> JobState:
        state = JobState.UNKNOWN
        tries = 0
        while True:
            try:
                state = await job.result_async(timeout)
                if state is JobState.COMPLETED:
                    return state
            except TimeoutError:
                state = JobState.TIMEOUT
            tries += 1

            if tries >= max_tries:
                break
            job = job.resubmit()

        if raise_error:
            raise RuntimeError(f'Job {job} failed with state {state}')
        else:
            return state

    return await asyncio.gather(*[wait_job(job) for job in jobs])
